from aiohttp import web
from joule.controllers import (
    root_controller,
    stream_controller,
    event_controller,
    folder_controller,
    data_controller,
    module_controller,
    app_controller,
    proxy_controller,
    master_controller,
    follower_controller,
    annotation_controller)

routes = [
    web.get('/', root_controller.index),
    web.get('/db/connection.json', root_controller.db_connection),
    web.get('/dbinfo', root_controller.dbinfo),
    web.get('/version', root_controller.version),
    web.get('/version.json', root_controller.version_json),
    # --- event stream routes ---
    web.get('/event.json', event_controller.info),
    web.put('/event/move.json', event_controller.move),
    web.put('/event.json', event_controller.update),
    web.post('/event.json', event_controller.create),
    web.delete('/event.json', event_controller.delete),
    # --- event stream data routes ---
    web.get('/event/data.json', event_controller.read_events),
    web.post('/event/data.json', event_controller.write_events),
    web.delete('/event/data.json', event_controller.remove_events),
    # --- stream routes ---
    web.get('/streams.json', folder_controller.index),  # legacy support for Rails API
    web.get('/stream.json', stream_controller.info),
    web.put('/stream/move.json', stream_controller.move),
    web.put('/stream.json', stream_controller.update),
    web.post('/stream.json', stream_controller.create),
    web.delete('/stream.json', stream_controller.delete),
    # --- folder routes ---
    web.get('/folders.json', folder_controller.index),
    web.get("/folder.json", folder_controller.info),
    web.put('/folder/move.json', folder_controller.move),
    web.put('/folder.json', folder_controller.update),
    web.delete('/folder.json', folder_controller.delete),
    # --- data routes ---
    web.get('/data', data_controller.read),
    web.get('/data.json', data_controller.read_json),
    web.get('/data/intervals.json', data_controller.intervals),
    web.post('/data/consolidate.json', data_controller.consolidate),
    web.post('/data', data_controller.write),
    web.delete('/data', data_controller.remove),
    # --- module routes ---
    web.get('/modules.json', module_controller.index),
    web.get('/module.json', module_controller.info),
    web.get('/module/logs.json', module_controller.logs),
    # --- app routes ---
    web.get('/app/auth', app_controller.auth),
    web.get('/app.json', app_controller.index),

    # --- proxy routes ---
    web.get('/proxies.json', proxy_controller.index),
    web.get('/proxy.json', proxy_controller.info),
    # -- master routes --
    web.get('/masters.json', master_controller.index),
    web.post('/master.json', master_controller.add),
    web.delete('/master.json', master_controller.delete),
    # -- follower routes --
    web.get('/followers.json', follower_controller.index),
    web.post('/follower.json', follower_controller.add),
    web.delete('/follower.json', follower_controller.delete),
    # -- annotation routes --
    web.get('/annotations/info.json', annotation_controller.info),
    web.get('/annotations.json', annotation_controller.index),
    web.put('/annotation.json', annotation_controller.update),
    web.post('/annotation.json', annotation_controller.create),
    web.delete('/annotation.json', annotation_controller.delete),
    web.delete('/stream/annotations.json', annotation_controller.delete_all)
]

insecure_routes = [
    ['POST', '/follower.json']
]



from typing import Union, Optional, List
import asyncio
import aiohttp
import numpy as np
from icecream import ic
from joule.models.pipes import compute_dtype

from .session import BaseSession
from .data_stream import (DataStream,
                          data_stream_get,
                          data_stream_create)

from joule.models.pipes import (Pipe,
                                InputPipe,
                                EmptyPipe,
                                LocalPipe,
                                interval_token)
from joule import errors

import logging

log = logging.getLogger('joule')


async def data_delete(session: BaseSession,
                      stream: Union[DataStream, str, int],
                      start: Optional[int] = None,
                      end: Optional[int] = None):
    params = {}
    if start is not None:
        params['start'] = int(start)
    if end is not None:
        params['end'] = int(end)
    if type(stream) is DataStream:
        params['id'] = stream.id
    elif type(stream) is str:
        params['path'] = stream
    elif type(stream) is int:
        params['id'] = stream
    else:
        raise errors.ApiError("Invalid stream datatype. Must be DataStream, Path, or ID")

    await session.delete("/data", params)


async def data_write(session: BaseSession,
                     stream: Union[DataStream, str, int],
                     start_time: Optional[int] = None,
                     end_time: Optional[int] = None,
                     ) -> Pipe:
    # stream must exist, does not automatically create a stream
    # retrieve the destination stream object
    dest_stream = await data_stream_get(session, stream)
    if start_time is not None or end_time is not None:
        await data_delete(session, dest_stream, start_time, end_time)

    if type(stream) is DataStream:
        if dest_stream.layout != stream.layout:
            raise errors.ApiError("DataStream [%s] configured for [%s] but destination is [%s]" % (
                stream.name, stream.layout, dest_stream.layout))
        # raise a warning if the element names do not match
        actual_names = [e.name for e in dest_stream.elements]
        requested_names = [e.name for e in stream.elements]
        if actual_names != requested_names:  # pragma: no cover
            log.warning("[%s] elements do not match the existing stream" % stream.name)

    # make sure the stream is not currently produced
    if dest_stream.is_destination:
        raise errors.ApiError("DataStream [%s] is already being produced" % dest_stream.name)

    # all checks passed, subscribe to the output
    async def close():
        await task

    pipe = LocalPipe(dest_stream.layout, name=dest_stream.name,
                     stream=dest_stream, close_cb=close, debug=False, write_limit=5)
    task = asyncio.create_task(_send_data(session, dest_stream, pipe))
    return pipe


async def data_intervals(session: BaseSession,
                         stream: Union[DataStream, str, int],
                         start_time: Optional[int] = None,
                         end_time: Optional[int] = None,
                         ) -> List:
    data = {}
    if type(stream) is DataStream:
        data["id"] = stream.id
    elif type(stream) is int:
        data["id"] = stream
    elif type(stream) is str:
        data["path"] = stream
    else:
        raise errors.ApiError("Invalid stream datatype. Must be DataStream, Path, or ID")
    if start_time is not None:
        data['start'] = int(start_time)
    if end_time is not None:
        data['end'] = int(end_time)
    return await session.get("/data/intervals.json", params=data)


async def data_consolidate(session: BaseSession,
                           stream: Union[DataStream, str, int],
                           start_time: Optional[int] = None,
                           end_time: Optional[int] = None,
                           max_gap: int = 2e6
                           ) -> List:
    data = {}
    if type(stream) is DataStream:
        data["id"] = stream.id
    elif type(stream) is int:
        data["id"] = stream
    elif type(stream) is str:
        data["path"] = stream
    else:
        raise errors.ApiError("Invalid stream datatype. Must be DataStream, Path, or ID")
    if start_time is not None:
        data['start'] = int(start_time)
    if end_time is not None:
        data['end'] = int(end_time)
    data['max_gap'] = int(max_gap)
    resp = await session.post("/data/consolidate.json", params=data)
    return resp['num_consolidated']


async def data_read(session: BaseSession,
                    stream: Union[DataStream, str, int],
                    start: Optional[int] = None,
                    end: Optional[int] = None,
                    max_rows: Optional[int] = None) -> Pipe:
    # make sure the input is compatible
    if type(stream) is not DataStream:
        stream = await data_stream_get(session, stream)
    # if type(stream) is DataStream:
    #    if src_stream.layout != src_stream.layout:
    #        raise errors.ApiError("Input [%s] configured for [%s] but source is [%s]" % (
    #            stream, stream.layout, src_stream.layout))
    # replace the stub stream (from config file) with actual stream
    pipe = LocalPipe(stream.layout, name=stream.name, stream=stream, write_limit=5)
    task = asyncio.create_task(_historic_reader(session,
                                                stream,
                                                pipe,
                                                start, end, max_rows))

    async def close():
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass

    pipe.close_cb = close
    return pipe


async def data_read_array(session: BaseSession,
                          stream: Union[DataStream, str, int],
                          start: Optional[int] = None,
                          end: Optional[int] = None,
                          max_rows: int = 10000,
                          flatten: bool = False) -> np.ndarray:
    """ Read the requested data into a numpy array, raises ValueError
    if the stream has more data than max_rows"""
    if type(stream) is not DataStream:
        stream = await data_stream_get(session, stream)
    # check if the data can be retrieved directly from NilmDB
    if await session.is_nilmdb_available():
        return await _read_nilmdb_data(session.nilmdb_url, stream, start, end, max_rows, flatten)
    else:
        pipe = await data_read(session, stream, start, end, max_rows)
        return await pipe.read_all(flatten=flatten)



async def data_subscribe(session: BaseSession,
                         stream: Union[DataStream, str, int]) -> Pipe:
    # make sure the input is compatible
    src_stream = await data_stream_get(session, stream)
    if type(stream) is DataStream:
        if src_stream.layout != src_stream.layout:
            raise errors.ApiError("Input [%s] configured for [%s] but source is [%s]" % (
                stream, stream.layout, src_stream.layout))

    # make sure the stream is being produced
    if not src_stream.is_destination:
        raise errors.ApiError(
            "DataStream [%s] is not being produced, specify time bounds for historic execution" % src_stream.name)
    # replace the stub stream (from config file) with actual stream
    # do not let the buffer grow beyond 5 server chunks
    pipe = LocalPipe(src_stream.layout, name=src_stream.name, stream=src_stream, write_limit=5)
    pipe.stream = src_stream

    task = asyncio.create_task(_live_reader(session,
                                            src_stream,
                                            pipe))

    async def close():
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass

    pipe.close_cb = close
    return pipe


async def _live_reader(session: BaseSession, my_stream: DataStream,
                       pipe_out: Pipe):
    log.info("requesting live connection to [%s]" % my_stream.name)
    params = {'id': my_stream.id, 'subscribe': '1'}
    try:
        raw_session = await session.get_session()
        async with raw_session.get(session.url + "/data",
                                   params=params,
                                   ssl=session.ssl_context) as response:
            if response.status != 200:  # pragma: no cover
                msg = await response.text()
                log.error("Error reading input [%s]: %s" % (my_stream.name, msg))
                await pipe_out.close()
                return
            pipe_out.change_layout(response.headers['joule-layout'])
            pipe_out.decimation_level = int(response.headers['joule-decimation'])
            pipe_in = InputPipe(layout=pipe_out.layout,
                                stream=my_stream, reader=response.content)
            while True:
                data = await pipe_in.read()
                pipe_in.consume(len(data))
                await pipe_out.write(data)
                if pipe_in.end_of_interval:
                    await pipe_out.close_interval()

    except (asyncio.CancelledError, EmptyPipe, aiohttp.ClientError):
        pass
    except Exception as e:
        print("unexpected exception: ", e)
        raise e
    await pipe_out.close()


async def _historic_reader(session: BaseSession,
                           my_stream: DataStream,
                           pipe_out: Pipe,
                           start_time: int, end_time: int,
                           max_rows: Optional[int]):
    log.info("requesting historic connection to [%s]" % my_stream.name)
    params = {'id': my_stream.id}
    if max_rows is not None:
        params['max-rows'] = max_rows
    if start_time is not None:
        params['start'] = int(start_time)
    if end_time is not None:
        params['end'] = int(end_time)
    try:
        my_session = await session.get_session()
        async with my_session.get(session.url + "/data",
                                  params=params,
                                  ssl=session.ssl_context) as response:
            if response.status != 200:  # pragma: no cover
                msg = await response.text()
                log.error("Error reading input [%s]: %s" % (my_stream.name, msg))
                await pipe_out.close()
                return
            if pipe_out.layout != response.headers['joule-layout']:
                print(f"WARNING: data pipe layout is incorrect, adjusting to {response.headers['joule-layout']}")
            pipe_out.change_layout(response.headers['joule-layout'])
            pipe_out.decimation_level = int(response.headers['joule-decimation'])
            pipe_in = InputPipe(layout=pipe_out.layout,
                                stream=my_stream, reader=response.content)
            while True:
                data = await pipe_in.read()
                pipe_in.consume(len(data))
                await pipe_out.write(data)
                if pipe_in.end_of_interval:
                    await pipe_out.close_interval()

    except (asyncio.CancelledError, EmptyPipe):
        pass
    except Exception as e:
        print("unexpected exception: ", e)
        raise e
    finally:
        await pipe_out.close()


async def _send_data(session: BaseSession,
                     stream: DataStream,
                     pipe: Pipe):
    # TODO is this necssary?
    output_complete = False

    async def _data_sender():
        nonlocal output_complete
        try:
            while True:
                data = await pipe.read()
                if len(data) > 0:
                    yield data.tobytes()
                if pipe.end_of_interval:
                    yield interval_token(stream.layout).tobytes()
                pipe.consume(len(data))
        except EmptyPipe:
            yield interval_token(stream.layout).tobytes()
        output_complete = True

    try:
        result = await session.post("/data",
                                    params={"id": stream.id},
                                    data=_data_sender(),
                                    chunked=True)
    except Exception as e:
        pipe.fail()
        raise e

async def _read_nilmdb_data(nilmdb_url, stream, start, end, max_rows, flatten) -> np.ndarray:
    url = f"{nilmdb_url}/stream/extract"
    params = {"path": f"/joule/{stream.id}",
              "binary": 1,
              "start": start,
              "end": end}
    dtype = compute_dtype(stream.layout)
    max_bytes = max_rows * dtype.itemsize

    data = b''
    async with aiohttp.ClientSession() as session:
        async with session.get(url, params=params) as resp:
            while not resp.content.at_eof() and len(data)<=max_bytes:
                data += await resp.content.read(max_bytes)
            if not resp.content.at_eof():
                print(f"WARNING: Requested time interval mas more data than {max_rows} rows of data")
    sdata = np.frombuffer(data[:max_bytes], dtype=dtype)
    if flatten:
        return np.c_[sdata['timestamp'][:, None], sdata['data']]
    else:
        return sdata



import click
import asyncio

from joule import errors
from joule.cli.config import Config, pass_config
from joule.utilities import timestamp_to_human, human_to_timestamp


@click.command(name="intervals")
@click.option('-s', "--start", help="timestamp or descriptive string")
@click.option('-e', "--end", help="timestamp or descriptive string")
@click.argument("stream")
@pass_config
def intervals(config: Config, start, end, stream: str):
    """Display time intervals with data."""
    if start is not None:
        try:
            start = human_to_timestamp(start)
        except ValueError:
            raise click.ClickException("invalid start time: [%s]" % start)
    if end is not None:
        try:
            end = human_to_timestamp(end)
        except ValueError:
            raise click.ClickException("invalid end time: [%s]" % end)

    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(_run(config.node, start, end, stream))

    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    except Exception as e:
        import traceback
        traceback.print_exc()
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


async def _run(node, start, end, stream):
    my_intervals = await node.data_intervals(stream, start, end)
    if len(my_intervals) == 0:
        print("no stream data")
        return
    for interval in my_intervals:
        print("[%s - %s]" % (
            timestamp_to_human(interval[0]),
            timestamp_to_human(interval[1])))



from aiohttp import web
from joule.models import DataStore
import joule


async def index(request: web.Request):
    return web.Response(text="Joule server")


async def db_connection(request: web.Request):
    return web.json_response(data=
                             request.app["module-connection-info"].to_json())


async def dbinfo(request: web.Request):
    data_store: DataStore = request.app["data-store"]
    data = await data_store.dbinfo()
    return web.json_response(data=data.to_json())


async def version_json(request: web.Request):
    return web.json_response(data={'version': joule.__version__,
                                   'name': request.app['name']})


async def version(request: web.Request):
    return web.Response(text=joule.__version__)



from joule.models.master import Master
import psycopg2

def update_users_from_file(users_file, db):
    print(f"Updating users based on [{users_file}]")
    with open(users_file, 'r') as f:
        for line in f:
            line = line.strip()
            if len(line) == 0 or line.startswith('#'):
                continue  # ignore comments and blank lines

            if line.startswith('DELETE'):
                parse_removal(line, db)
            else:
                parse_addition(line, db)


def parse_removal(line, db):
    params = line.split(' ')
    if len(params) == 2:
        if params[0] != 'DELETE':
            print(f"ERROR: invalid delete syntax in users file: {line}")
            return
        remove_user(params[1], db)
    elif len(params) == 3:
        if params[0] != 'DELETE' or params[1] != 'LIKE':
            print(f"ERROR: invalid delete syntax in users file: {line}")
            return
        remove_user_like(params[2], db)
    else:
        print(f"ERROR: invalid delete syntax in users file: {line}")


def parse_addition(line, db):
    params = line.split(',')
    if len(params) != 2:
        print(f"ERROR: invalid syntax in users file: {line}")
        return
    user, key = params
    if len(key) < 32:
        print(f"ERROR: key is too short, must be >= 32 characters: {line}")
        return
    add_user(user.strip(), key.strip(), db)


def remove_user(user, db):
    users = db.query(Master). \
        filter(Master.name == user)
    for user in users:
        print(f"Removing user {user.name}")
        db.delete(user)
    db.commit()


def remove_user_like(name, db):
    users = db.query(Master). \
        filter(Master.name.like(name))
    for user in users:
        print(f"Removing user {user.name}")
        db.delete(user)
    db.commit()


def add_user(name, key, db):
    # make sure the key does not exist already
    if db.query(Master).filter(Master.key==key).first():
        print(f"Cannot add {name}, the requested key is already associated with another user")
        return
    # check if this name already exists
    master = db.query(Master). \
        filter(Master.type == Master.TYPE.USER). \
        filter(Master.name == name).first()
    if master is not None:
        if master.key == key:
            return  # nothing to do
        master.key = key
        print(f"Modifying API key for {name}")
    else:
        # create a new name
        master = Master(name=name,
                        type=Master.TYPE.USER,
                        key=key)
        master.grantor_id = None  # not granted by another name
        print(f"Adding user {name}")
    # save the changes
    db.add(master)
    db.commit()



import asyncio
import numpy as np
import logging
import argparse
import sys

from joule.client.base_module import BaseModule
from joule.models.pipes import Pipe
from joule import errors


class ReaderModule(BaseModule):
    """
    Inherit from this class and implement a :meth:`run` coroutine to create a Joule reader module.
    Other methods documented below may be implemented as desired.
    """

    async def setup(self, parsed_args, app, output):
        """
        Configure the module, executes before :meth:`run`

        Args:
            parsed_args:
            app:
            output:

        """
        pass

    async def run(self, parsed_args: argparse.Namespace, output: Pipe):
        """
        This method must be implemented. It should run in a loop, if it returns the module
        stops.

        Args:
            parsed_args: command line arguments, configure with :meth:`custom_args`
            output: pipe connection to the output data stream

        .. code-block:: python

            class ModuleDemo(ReaderModule):

                def run(self, parsed_args, output):
                     while(not self.stop_requested):
                        data = self.read_sensor()
                        await output.write(data)

                def self.read_sensor(self) -> np.ndarray:
                    # custom logic specific to the reader

                #... other module code

        """
        assert False, "implement in child class"  # pragma: no cover

    async def run_as_task(self, parsed_args, app):
        # check if we should use stdout (no fd's and no configs)
        if (parsed_args.pipes == "unset" and
                parsed_args.module_config == "unset"):
            output = StdoutPipe()
        else:
            try:
                (pipes_in, pipes_out) = await self._build_pipes(parsed_args)
            except errors.ApiError as e:
                logging.error(str(e))
                return asyncio.create_task(asyncio.sleep(0))
            if len(pipes_out.keys()) != 1:
                logging.error("Reader module must have a single output")
                return asyncio.create_task(asyncio.sleep(0))
            output = list(pipes_out.values())[0]
        await self.setup(parsed_args, app, output)
        return asyncio.create_task(self.run(parsed_args, output))


class StdoutPipe:
    @staticmethod
    async def write(data: np.ndarray):
        # check if this is a structured array, if so flatten it
        if data.ndim == 1:
            data = np.c_[data['timestamp'][:, None], data['data']]
        for row in data:
            ts = row[0]
            vals = row[1:]
            print("%d %s" % (ts, " ".join([repr(x) for x in vals])))

    async def close_interval(self):
        print("--- interval gap ---", file=sys.stderr)

    async def close(self):
        pass



import click
import asyncio

from joule import errors
from joule.api import node
from joule.api.proxy import (proxy_get)
from joule.cli.config import Config, pass_config


@click.command(name="info")
@click.argument("name")
@pass_config
def cli_info(config: Config, name: str):
    """Display proxy information"""
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            _run(config.node, name))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


async def _run(node: node.BaseNode, name: str):
    proxy = await node.proxy_get(name)
    # display proxy information
    click.echo()
    click.echo("ID:\n\t%d" % proxy.id)
    click.echo("Name:\n\t%s" % proxy.name)
    click.echo("Proxy URL:\n\t%s/interface/p%d/" % (node.url, proxy.id))
    click.echo("Source URL:\n\t%s" % proxy.url)






import os
import configparser
import ipaddress
import psycopg2
import yarl
import ssl
import aiohttp
import asyncio

from joule.models import config, Proxy
from joule.errors import ConfigurationError

"""
# Example configuration (including optional lines)

[Main]
  ModuleDirectory = /etc/joule/module_configs
  StreamDirectory = /etc/joule/stream_configs
  IPAddress = 127.0.0.1
  Port = 8088
  Database = joule:joule@localhost:5432/joule
  InsertPeriod = 5
  CleanupPeriod = 60
  MaxLogLines = 100
  NilmdbUrl = http://localhost/nilmdb
[Security]
   Certificate =  /etc/joule/security/server.crt
   Key = /etc/joule/security/server.key
   CertificateAuthority = /etc/joule/security/ca.crt
[Proxies]
  site1 = http://localhost:3000
  site2 = https://othersite.com
"""


def run(custom_values=None, verify=True) -> config.JouleConfig:
    """provide a dict INI configuration to override defaults
       if verify is True, perform checks on settings to make sure they are appropriate"""
    my_configs = configparser.ConfigParser()
    my_configs.read_dict(config.DEFAULT_CONFIG)
    if custom_values is not None:
        my_configs.read_dict(custom_values)

    main_config = my_configs['Main']
    # Node name
    node_name = main_config['Name']

    # ModuleDirectory
    module_directory = main_config['ModuleDirectory']
    if not os.path.isdir(module_directory) and verify:
        raise ConfigurationError(
            "ModuleDirectory [%s] does not exist" % module_directory)
    # StreamDirectory
    stream_directory = main_config['StreamDirectory']
    if not os.path.isdir(stream_directory) and verify:
        raise ConfigurationError(
            "StreamDirectory [%s] does not exist" % stream_directory)

    # Specify IPAddress and Port to listen on network interface
    # IPAddress
    if 'IPAddress' in main_config:
        ip_address = main_config['IPAddress']
        try:
            ipaddress.ip_address(ip_address)
        except ValueError as e:
            raise ConfigurationError("IPAddress is invalid") from e
        # Port
        try:
            port = int(main_config['Port'])
            if port < 0 or port > 65535:
                raise ValueError()
        except ValueError as e:
            raise ConfigurationError("Port must be between 0 - 65535") from e
    else:
        port = None
        ip_address = None

    # SocketDirectory
    socket_directory = main_config['SocketDirectory']
    if not os.path.isdir(socket_directory) and verify:
        try:
            os.mkdir(socket_directory)
            # make sure the ownership is correct
            os.chmod(socket_directory, 0o700)
        except FileExistsError:
            raise ConfigurationError("SocketDirectory [%s] is a file" % socket_directory)
        except PermissionError:
            raise ConfigurationError("Cannot create SocketDirectory at [%s]" % socket_directory)

    if not os.access(socket_directory, os.W_OK) and verify:
        raise ConfigurationError(
            "SocketDirectory [%s] is not writable" % socket_directory)

    # Nilmdb URL
    if 'NilmdbUrl' in main_config and main_config['NilmdbUrl'] != '':
        nilmdb_url = main_config['NilmdbUrl']
        if verify:
            loop = asyncio.get_event_loop()
            loop.run_until_complete(verify_nilmdb_url(nilmdb_url))

    else:
        nilmdb_url = None

    # Database
    if 'Database' in main_config:
        database = "postgresql://" + main_config['Database']
    elif verify:
        raise ConfigurationError("Missing [Database] configuration")
    else:  # pragma: no cover
        database = ''  # this is invalid of course, just used in unit testing
    if verify:
        # check to see if this is a valid database DSN
        try:
            conn = psycopg2.connect(database)
            conn.close()
        except psycopg2.Error:
            raise ConfigurationError("Cannot connect to database [%s]" % database)

    # InsertPeriod
    try:
        insert_period = int(main_config['InsertPeriod'])
        if insert_period <= 0:
            raise ValueError()
    except ValueError:
        raise ConfigurationError("InsertPeriod must be a postive number")

    # CleanupPeriod
    try:
        cleanup_period = int(main_config['CleanupPeriod'])
        if cleanup_period <= 0 or cleanup_period < insert_period:
            raise ValueError()
    except ValueError:
        raise ConfigurationError("CleanupPeriod must be a postive number > InsertPeriod")

    # Max Log Lines
    try:
        max_log_lines = int(main_config['MaxLogLines'])
        if max_log_lines <= 0:
            raise ValueError()
    except ValueError:
        raise ConfigurationError("MaxLogLines must be a postive number")

    # Authorized Users File
    users_file = None
    if 'UsersFile' in main_config:
        users_file = main_config['UsersFile']
        if not os.path.isfile(users_file) and verify:
            raise ConfigurationError(
                "UsersFile [%s] does not exist" % users_file)

    # Security configuration
    if 'Security' in my_configs:
        security = config.SecurityConfig(my_configs['Security']['Certificate'],
                                         my_configs['Security']['Key'],
                                         my_configs.get('Security',
                                                        'CertificateAuthority', fallback=""))
    else:
        security = None

    # Proxies
    uuid = 0
    proxies = []
    if 'Proxies' in my_configs:
        for name in my_configs['Proxies']:
            url_str = my_configs['Proxies'][name]
            # make sure proxy url ends with /
            if url_str[-1] != '/':
                url_str += '/'
            url = yarl.URL(url_str)
            proxies.append(Proxy(name, uuid, url))
            uuid += 1

    return config.JouleConfig(
        name=node_name,
        module_directory=module_directory,
        stream_directory=stream_directory,
        ip_address=ip_address,
        port=port,
        socket_directory=socket_directory,
        database=database,
        insert_period=insert_period,
        cleanup_period=cleanup_period,
        max_log_lines=max_log_lines,
        nilmdb_url=nilmdb_url,
        proxies=proxies,
        security=security,
        users_file=users_file
    )


async def verify_nilmdb_url(url):
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as resp:
                text = await resp.text()
                if 'NilmDB' not in text:
                    raise ConfigurationError(
                        "Host at [%s] is not a NilmDB server" % url
                    )
    except aiohttp.ClientError:
        raise ConfigurationError(
            "Cannot contact NilmDB at [%s]" % url
        )



import click
from .list import cli_list
from .delete import follower_delete


@click.group(name="follower")
def follower():
    """Manage node followers."""
    pass  # pragma: no cover


follower.add_command(cli_list)
follower.add_command(follower_delete)



from joule import errors
from joule.utilities import connection_info
import aiohttp

class BaseSession:

    def __init__(self):
        self.url = ""
        self._session = None
        self.ssl_context = None
        self.cafile = ""
        self._nilmdb_checked = False
        self._nilmdb_available = False
        self._nilmdb_url = ""

    async def get_session(self):
        return self._session

    async def get(self, path, params=None):
        return await self._request("GET", path, params=params)

    async def post(self, path, json=None, params=None, data=None, chunked=None):
        return await self._request("POST", path, json=json,
                                   params=params, data=data, chunked=chunked)

    async def put(self, path, json):
        return await self._request("PUT", path, json=json)

    async def delete(self, path, params):
        return await self._request("DELETE", path, params=params)

    async def _request(self, method, path, data=None, json=None, params=None, chunked=None):
        raise errors.ApiError("Implement in child class")

    async def close(self):
        if self._session is not None:
            await self._session.close()
        self._session = None

    async def is_nilmdb_available(self):
        # return cached result if available
        if self._nilmdb_checked:
            return self._nilmdb_available
        # this is the first time, we have to check...
        self._nilmdb_checked=True
        resp = await self.get("/db/connection.json")
        conn_info = connection_info.from_json(resp)
        if conn_info.nilmdb_url is None or conn_info.nilmdb_url == "":
            self._nilmdb_available = False
            return False # This Joule node does not use NilmDB or is out of date
        self._nilmdb_url = conn_info.nilmdb_url
        try:
            async with aiohttp.ClientSession() as session:
                async with session.get(f'{self._nilmdb_url}') as response:
                    if response.status != 200:
                        raise Exception("HTTP error")
                    html = await response.text()
                    if "NilmDB" not in html:
                        raise Exception("Not a NilmDB server")
        except Exception as e:
            self._nilmdb_available = False
            return False # something went wrong with the request
        self._nilmdb_available = True
        return True # all checks passed, NilmDB is available

    def set_nilmdb_url(self, nilmdb_url):
        """Override auto check, specify nilmdb_url manually"""
        self._nilmdb_checked = True
        self._nilmdb_available = True
        self._nilmdb_url = nilmdb_url

    @property
    def nilmdb_url(self):
        if not self._nilmdb_checked:
            raise Exception("Must call is_nilmdb_available() first")
        if not self._nilmdb_available:
            raise Exception("NilmDB backend is not available")
        return self._nilmdb_url


import click
import asyncio

from joule import errors
from joule.api import BaseNode
from joule.cli.config import pass_config


@click.command(name="logs")
@click.argument("name")
@pass_config
def cli_logs(config, name):
    """Display a module's log output"""
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            _run(config.node, name))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


async def _run(node: BaseNode, name:str):
    logs = await node.module_logs(name)
    for line in logs:
        print(line)



import asyncio
import numpy as np
import logging

from joule.models.pipes import Pipe, interval_token
from joule.models.pipes.errors import PipeError
from icecream import ic
log = logging.getLogger('joule')


class OutputPipe(Pipe):

    def __init__(self, name=None, stream=None, layout=None, close_cb=None,
                 writer: asyncio.StreamWriter=None, writer_factory=None):
        super().__init__(name=name, stream=stream, layout=layout,
                         direction=Pipe.DIRECTION.OUTPUT)
        self.writer_factory = writer_factory
        self.writer: asyncio.StreamWriter = writer
        self.close_cb = close_cb
        # caching
        self._caching = False
        self._cache_index = 0
        self._cache = None

    async def write(self, data):
        if self.closed:
            raise PipeError("Cannot write to a closed pipe")
        if not self._validate_data(data):
            return
        # make sure dtype is structured
        sdata = self._apply_dtype(data)

        if self._caching:
            for row in sdata:
                self._cache[self._cache_index] = row
                self._cache_index += 1
                if self._cache_index >= len(self._cache):
                    await self.flush_cache()
        else:
            await self._write(sdata)

    def enable_cache(self, lines: int):
        self._caching = True
        self._cache = np.empty(lines, self.dtype)
        self._cache_index = 0

    async def flush_cache(self):
        if self.closed:
            raise PipeError("Pipe is closed")
        if self._cache_index > 0:
            await self._write(self._cache[:self._cache_index])
            self._cache_index = 0
            self._cache = np.empty(len(self._cache), self.dtype)

    async def _write(self, sdata):
        if self.writer is None:
            self.writer = await self.writer_factory()

        # send data to subscribers
        for pipe in self.subscribers:
            await pipe.write(sdata)

        # send data out
        self.writer.write(sdata.tobytes())
        await self.writer.drain()
        await asyncio.sleep(0)

    async def close_interval(self):
        if self.closed:
            raise PipeError("Pipe is closed")
        if self.writer is None:
            return  # nothing has been written yet so nothing to close
        if self._caching:
            await self.flush_cache()
        self.writer.write(interval_token(self.layout).tobytes())
        await self.writer.drain()

    def close_interval_nowait(self):
        if self.closed:
            raise PipeError("Pipe is closed")
        if self.writer is None:
            return  # nothing has been written yet so nothing to close
        if self._cache_index > 0:
            log.warning("dumping %d rows of cached data due on %s" % (self._cache_index, self.name))
            self._cache = np.empty(len(self._cache), self.dtype)
            self._cache_index = 0
        self.writer.write(interval_token(self.layout).tobytes())

    async def close(self):
        if self.closed:
            return
        # if the cache is enabled, flush it
        if self._caching:
            await self.flush_cache()
        if self.close_cb is not None:
            await self.close_cb()
        if self.writer is not None:
            self.writer.close()
            # available in python3.7
            await self.writer.wait_closed()
            # Hack to close the transport (if wait_closed is not available)
            # await asyncio.sleep(0.01)
            self.writer = None
        # close any subscribers
        for pipe in self.subscribers:
            await pipe.close()
        self.closed = True




from typing import Dict
from joule.errors import ApiError


class ConnectionInfo:
    """
    Encapsulates database connection information
    """

    def __init__(self, username: str, password: str, port: int, database: str, host: str, nilmdb_url):
        """
        Attributes:
            username (str): database username
            password (str): database password
            port (int): database port
            database (str): database name
            host (str): hostname
            nilmdb_url (str): NilmDB URL (empty if unused)

        """
        self.username = username
        self.password = password
        self.port = port
        self.database = database
        self.host = host
        self.nilmdb_url = nilmdb_url

    def to_json(self) -> Dict:
        """
        Convert connection parameters to a dictionary of attributes appropriate
        for transmission as JSON

        """
        return {
            "username": self.username,
            "password": self.password,
            "port": self.port,
            "database": self.database,
            "host": self.host,
            "nilmdb_url": self.nilmdb_url
        }

    def to_dsn(self) -> str:
        """
        Convert connection parameters to a DSN string used by SQLAlchemy and other
        database engines.

        """
        return "postgresql://%s:%s@%s:%s/%s" % (self.username, self.password,
                                                self.host, self.port, self.database)


def from_json(data: Dict) -> 'ConnectionInfo':
    # keep client backwards compatible with earlier jouled versions
    if "nilmdb_url" not in data:
        data["nilmdb_url"] = ""
    try:
        info = ConnectionInfo(username=data["username"],
                              password=data["password"],
                              port=data["port"],
                              database=data["database"],
                              host=data["host"],
                              nilmdb_url=data["nilmdb_url"])
    except KeyError as e:
        raise ApiError("invalid connection info, missing [%s]" % str(e))
    return info



import asyncpg
import json

from typing import List, Optional, TYPE_CHECKING, Dict, Callable, Tuple
from joule.models.data_store import psql_helpers
import joule.utilities

if TYPE_CHECKING:
    from joule.models import EventStream


class EventStore:

    def __init__(self, dsn=""):
        self.dsn = dsn
        self.pool = None

    async def initialize(self, pool=None) -> None:
        if pool is None:
            if self.dsn == "":
                raise ValueError("Must specify either dsn or pool")
            self.pool = await asyncpg.create_pool(self.dsn, command_timeout=60)
        else:
            self.pool = pool
        async with self.pool.acquire() as conn:
            await psql_helpers.create_event_table(conn)

    async def upsert(self, stream: 'EventStream', events: List):
        updated_events = list(filter(lambda event: "id" in event and event["id"] is not None, events))
        new_events = list(filter(lambda event: "id" not in event or event["id"] is None, events))
        async with self.pool.acquire() as conn:
            if len(new_events) > 0:
                seq = "'data.events_id_seq'"
                query = await conn.fetch(
                    f"select setval({seq},nextval({seq}) + {len(new_events)} -1) as stop")

                serial = query[0]["stop"] - len(new_events)

                for e in new_events:
                    serial += 1
                    e["id"] = serial
                assert serial == query[0]["stop"]
                mapper = map(event_to_record_mapper(stream), new_events)
                await conn.copy_records_to_table("events",
                                                 columns=['id', 'time', 'end_time', 'event_stream_id', 'content'],
                                                 schema_name='data',
                                                 records=mapper)
            if len(updated_events) > 0:
                mapper = map(event_to_record_mapper(stream), updated_events)
                for e in mapper:
                    await conn.execute("UPDATE data.events SET time=$2, "
                                       "end_time=$3, event_stream_id=$4, "
                                       "content=$5 WHERE id=$1",
                                       *e)
        return events

    async def count(self, stream: 'EventStream',
                    start: Optional[int] = None, end: Optional[int] = None,
                    json_filter=None) -> int:
        if end is not None and start is not None and end <= start:
            raise ValueError("Invalid time bounds start [%d] must be < end [%d]" % (start, end))
        query = "SELECT count(*) FROM data.events "
        where_clause = psql_helpers.query_time_bounds(start, end,
                                                      start_col_name='time',
                                                      end_col_name='end_time')
        if len(where_clause) == 0:
            where_clause = "WHERE "
        else:
            where_clause += " AND "
        where_clause += "event_stream_id=%d" % stream.id
        if json_filter is not None and len(json_filter) > 0:
            where_clause += " AND " + psql_helpers.query_event_json(json_filter)
        query += where_clause
        async with self.pool.acquire() as conn:
            record = await conn.fetch(query)
            return record[0]['count']

    async def extract(self, stream: 'EventStream',
                      start: Optional[int] = None,
                      end: Optional[int] = None,
                      json_filter=None,
                      limit=None) -> List[Dict]:
        if end is not None and start is not None and end <= start:
            raise ValueError("Invalid time bounds start [%d] must be < end [%d]" % (start, end))
        query = "SELECT id, time, end_time, content FROM data.events "
        where_clause = psql_helpers.query_time_bounds(start, end,
                                                      start_col_name='time',
                                                      end_col_name='end_time')
        if len(where_clause) == 0:
            where_clause = "WHERE "
        else:
            where_clause += " AND "
        where_clause += "event_stream_id=%d" % stream.id
        if json_filter is not None and len(json_filter) > 0:
            where_clause += " AND " + psql_helpers.query_event_json(json_filter)
        query += where_clause
        if limit is not None:
            assert limit > 0, "limit must be > 0"
            if start is None and end is not None:
                query += " ORDER BY time DESC"
            else:
                query += " ORDER BY time ASC"
            query += f" LIMIT {limit}"
        else:
            query += " ORDER BY time ASC"
        async with self.pool.acquire() as conn:
            records = await conn.fetch(query)
            events = list(map(record_to_event, records))
            events.sort(key=lambda e: e["start_time"])
            return events

    async def remove(self, stream: 'EventStream', start: Optional[int] = None,
                     end: Optional[int] = None,
                     json_filter=None):
        query = "DELETE FROM data.events "
        where_clause = psql_helpers.query_time_bounds(start, end,
                                                      start_col_name='time',
                                                      end_col_name='end_time')
        if len(where_clause) == 0:
            where_clause = "WHERE "
        else:
            where_clause += " AND "
        where_clause += "event_stream_id=%d" % stream.id
        if json_filter is not None and len(json_filter) > 0:
            where_clause += " AND " + psql_helpers.query_event_json(json_filter)
        query += where_clause
        async with self.pool.acquire() as conn:
            return await conn.execute(query)

    async def destroy(self, stream: 'EventStream'):
        await self.remove(stream)

    async def info(self, streams: List['EventStream']) -> Dict[int, 'StreamInfo']:
        if len(streams) == 0:
            return {}  # no event streams
        info_objects = {s.id: StreamInfo(None, None, 0, 0, 0) for s in streams}
        async with self.pool.acquire() as conn:
            query = "select count(*), event_stream_id, max(time) as end_time," \
                    " min(time) as start_time from data.events "
            if len(streams) < 10:
                stream_ids = ','.join([str(s.id) for s in streams])
                query += f"where event_stream_id in ({stream_ids}) "
            query += "group by event_stream_id"
            rows = await conn.fetch(query)
            for row in rows:
                if row['event_stream_id'] in info_objects.keys():
                    end_time = joule.utilities.datetime_to_timestamp(row['end_time'])
                    start_time = joule.utilities.datetime_to_timestamp(row['start_time'])
                    total_time = end_time - start_time
                    total_bytes = 0  # TODO
                    info_objects[row['event_stream_id']] = StreamInfo(start_time, end_time, row['count'],
                                                                      total_time, total_bytes)
        return info_objects

    async def close(self):
        await self.pool.close()


class StreamInfo:
    def __init__(self, start: Optional[int], end: Optional[int], event_count: int,
                 total_time: int = 0, bytes: int = 0):
        self.start = start
        self.end = end
        self.event_count = event_count
        self.bytes = bytes
        self.total_time = total_time

    def __repr__(self):
        return "<DataStreamInfo start=%r end=%r event_count=%r, total_time=%r>" % (
            self.start, self.end, self.event_count, self.total_time)

    def __eq__(self, other):
        return self.start == other.start and \
               self.end == other.end and \
               self.event_count == other.event_count and \
               self.bytes == other.bytes and \
               self.total_time == other.total_time

    def to_json(self):
        return {
            "start": self.start,
            "end": self.end,
            "event_count": self.event_count,
            "bytes": self.bytes,
            "total_time": self.total_time
        }


def event_to_record_mapper(stream: 'EventStream') -> Callable[[Dict], Tuple]:
    stream_id = stream.id

    def mapper(event: Dict) -> Tuple:
        start = joule.utilities.timestamp_to_datetime(event['start_time'])
        if event['end_time'] is not None:
            if event['start_time'] >= event['end_time']:
                raise ValueError("Event end [%d] cannot be before start [%d]" % (
                    event['start_time'], event['end_time']
                ))
            end = joule.utilities.timestamp_to_datetime(event['end_time'])
        else:
            end = None
        return event["id"], start, end, stream_id, json.dumps(event['content'])

    return mapper


def record_to_event(record: asyncpg.Record) -> Dict:
    if record['end_time'] is not None:
        end = joule.utilities.datetime_to_timestamp(record['end_time'])
    else:
        end = None
    return {
        'id': record['id'],
        'start_time': joule.utilities.datetime_to_timestamp(record['time']),
        'end_time': end,
        'content': json.loads(record['content'])}



import datetime
import dateparser


# --------- Utility functions from Jim Paris ------------


def time_now() -> int:
    """
    :return:     current time in UNIX microseconds
    """
    return int(datetime.datetime.now().timestamp() * 1e6)


# Range
min_timestamp = (-2 ** 63)
max_timestamp = (2 ** 63 - 1)


def timestamp_to_human(timestamp: int) -> str:
    """Convert a timestamp (integer microseconds since epoch) to a
    human-readable string, using the local timezone for display
    (e.g. from the TZ env var)."""
    if timestamp == min_timestamp:
        return "(minimum)"
    if timestamp == max_timestamp:
        return "(maximum)"
    dt = datetime.datetime.fromtimestamp(timestamp_to_unix(timestamp))
    return dt.strftime("%a, %d %b %Y %H:%M:%S %z")


def human_to_timestamp(time: str) -> int:
    """Convert a time specification into a UNIX microsecond timestamp. Time specification
        may be a wide variety of date formats, relative interval such as "one minute ago", or a numeric
        timestamp. Raises :exc:ValueError for invalid time specification"""
    try:
        return int(time)
    except ValueError:
        pass
    time = dateparser.parse(time)
    if time is None:
        raise ValueError
    return int(time.timestamp() * 1e6)


def unix_to_timestamp(unix):
    """Convert a Unix timestamp (floating point seconds since epoch)
    into a NILM timestamp (integer microseconds since epoch)"""
    return int(round(unix * 1e6))


def timestamp_to_datetime(timestamp: int):
    return datetime.datetime.utcfromtimestamp(timestamp / 1e6)


def datetime_to_timestamp(date: datetime):
    return int(date.replace(tzinfo=datetime.timezone.utc).timestamp() * 1e6)


seconds_to_timestamp = unix_to_timestamp


def timestamp_to_unix(timestamp):
    """Convert a NILM timestamp (integer microseconds since epoch)
    into a Unix timestamp (floating point seconds since epoch)"""
    return timestamp / 1e6


timestamp_to_seconds = timestamp_to_unix



from sqlalchemy import Column, Integer, Enum, String, ForeignKey
import enum
import secrets

from joule.models.meta import Base


class Master(Base):
    """
    Attributes:
        name (str): master name
        type (Master.TYPE): whether the master is a user, joule, or lumen
        key (str): API key for accessing this node
    """
    __tablename__ = 'master'
    __table_args__ = {"schema": "metadata"}

    id: int = Column(Integer, primary_key=True)
    name: str = Column(String, nullable=False)
    key: str = Column(String, nullable=False, unique=True)

    class TYPE(enum.Enum):
        USER = enum.auto()
        JOULE_NODE = enum.auto()
        LUMEN_NODE = enum.auto()

    type: TYPE = Column(Enum(TYPE))
    grantor_id: int = Column(Integer, ForeignKey('metadata.master.id'))

    def __repr__(self):
        return "<Master(id=%r name=%r, type=%r, grantor_id=%r)>" % (
            self.id, self.name, self.type, self.grantor_id)

    def to_json(self):
        """
        Returns: Dictionary of Master attributes
        """
        return {
            'id': self.id,
            'name': self.name,
            'type': self.type.name,
            'grantor_id': self.grantor_id
        }


def make_key():
    # create an API key
    return secrets.token_urlsafe(32)



import scipy.signal
import asyncio
import textwrap
import numpy as np

import joule


ARGS_DESC = """
---
:name:
  Median Filter
:author:
  John Donnal
:license:
  Open
:url:
  http://git.wattsworth.net/wattsworth/joule.git
:usage:
    Apply a windowed median filter to the input stream.
    
    | Arguments | Description                     |
    |-----------|---------------------------------|
    |``window`` | samples to average, must be odd |


:inputs:
  input
  :  **<any type>** with N elements

:outputs:
  output
  :  **float32** with N elements

:stream_configs:
  #input#
     [Main]
     name = Raw Data
     path = /path/to/input
     datatype = int32
     keep = 1w

     [Element1]
     name = Element 1

     [Element2]
     name = Element 2

     #additional elements...

  #output#
     [Main]
     name = Filtered Data
     path = /path/to/output
     datatype = float32
     keep = 1w

     #same number of elements as input
     [Element1]
     name = Element 1

     [Element2]
     name = Element 2

     #additional elements...

:module_config:
    [Main]
    name = Mean Filter
    exec_cmd = joule-median-filter

    [Arguments]
    # must be odd
    window = 11

    [Inputs]
    input = /path/to/input

    [Outputs]
    output = /path/to/output
---
"""


class MedianFilter(joule.FilterModule):
    """Compute the median of the input"""

    def custom_args(self, parser):  # pragma: no cover
        grp = parser.add_argument_group("module",
                                        "module specific arguments")
        grp.add_argument("--window", type=int, required=True,
                         help="window length")
        parser.description = textwrap.dedent(ARGS_DESC)

    async def run(self, parsed_args, inputs, outputs):
        N = parsed_args.window
        stream_in = inputs["input"]
        stream_out = outputs["output"]
        while not self.stop_requested:
            sarray_in = await stream_in.read()
            # not enough data, wait for more
            if len(sarray_in) < (N * 2):
                # check if the pipe is closed
                if stream_in.closed:  # pragma: no cover
                    return
                # check if this is the end of an interval
                # if so we can't use this data so discard it
                if stream_in.end_of_interval:
                    stream_in.consume(len(sarray_in))
                    await stream_out.close_interval()
                await asyncio.sleep(0.1)
                continue
            # allocate output array
            output_len = len(sarray_in)-N+1
            sarray_out = np.zeros(output_len, dtype=stream_out.dtype)
            filtered = scipy.signal.medfilt(sarray_in['data'], [N, 1])
            bound = int(N / 2)
            sarray_out['data'] = filtered[bound:-bound]
            sarray_out['timestamp'] = sarray_in['timestamp'][bound:-bound]
            await stream_out.write(sarray_out)
            # hard to isolate in test, usually hits line 109
            if stream_in.end_of_interval:  # pragma: no cover
                await stream_out.close_interval()
                # dump all of the data because we don't
                # want to mix median across intervals
                stream_in.consume(len(sarray_in))
            else:
                stream_in.consume(len(sarray_out))


def main():  # pragma: no cover
    r = MedianFilter()
    r.start()

    
if __name__ == "__main__":
    main()



import click
import asyncio

from joule import errors
from joule.cli.config import pass_config


@click.command(name="delete")
@click.argument("stream")
@pass_config
def cli_delete(config, stream):
    """Delete a data stream"""
    # make sure the stream exists
    asyncio.run(run(config, stream))


async def run(config, stream):
    stream_type = None
    try:
        await config.node.data_stream_get(stream)
        stream_type = "data"
    except errors.ApiError:
        pass
    try:
        if stream_type is None:
            await config.node.event_stream_get(stream)
            stream_type = "event"
    except errors.ApiError:
        pass
    if stream_type is None:
        await config.close_node()
        raise click.ClickException("Stream does not exist")

    if not click.confirm("Delete %s stream [%s]?" % (stream_type, stream)):
        click.echo("Cancelled")
        await config.close_node()
        return

    try:
        if stream_type == 'data':
            await config.node.data_stream_delete(stream)
        elif stream_type == 'event':
            await config.node.event_stream_delete(stream)
        else:
            raise click.ClickException("unknown stream type [%s]" % stream_type)
        click.echo("OK")

    except errors.ApiError as e:
        raise click.ClickException(str(e))
    finally:
        await config.close_node()



from typing import Optional, List
from asyncio import AbstractEventLoop, get_event_loop

from .base_node import BaseNode
from .node_config import NodeConfig
from joule.api.session import TcpSession


class TcpNode(BaseNode):
    def __init__(self, name: str, url: str, key: str, cafile: str = ""):
        session = TcpSession(url, key, cafile)
        self._url = url
        self._key = key
        super().__init__(name, session)
        self.url = url

    def __repr__(self):
        return "<joule.api.node.TcpNode url=\"%s\">" % self.url

    def to_config(self) -> NodeConfig:
        return NodeConfig(self.name, self.url, self._key)

    async def follower_list(self) -> List[BaseNode]:
        resp = await self.session.get("/followers.json")
        return [TcpNode(item['name'], item['location'], item['key'], self.session.cafile) for item in resp]



import click
import asyncio

from joule.cli.config import pass_config
from joule import errors


@click.command(name="info")
@pass_config
def node_info(config):
    """Display information about a node."""
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            _run(config.node))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e

    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


# https://stackoverflow.com/questions/1094841
def sizeof_fmt(num, suffix='B'):
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


async def _run(node):
    info = await node.info()
    click.echo("Server Version:   \t%s" % info.version)
    click.echo("Database Location:\t%s" % info.path)
    if info.path != "--remote-database--":
        click.echo("Database Size:    \t%s" % sizeof_fmt(info.size_db))
        click.echo("Space Available:  \t%s" % sizeof_fmt(info.size_free))



import ssl
import aiohttp
import asyncio
from .base_session import BaseSession
from joule import errors


# import logging
# logging.basicConfig(filename='/home/vagrant/app.log', filemode='w', format='%(name)s - %(levelname)s - %(message)s')

class TcpSession(BaseSession):

    def __init__(self, url: str, key: str, cafile: str):
        super().__init__()
        self.url = url
        self.key = key
        self.ssl_context = None
        self.cafile = cafile
        # for https nodes
        if self.url.startswith("https"):
            self.ssl_context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
            # load cafile to verify the node
            if cafile != "":
                self.ssl_context.load_verify_locations(cafile=cafile)
            else:
                self.ssl_context.check_hostname = False
                self.ssl_context.verify_mode = ssl.CERT_NONE

    def __repr__(self):
        return "<joule.api.session.TcpSession url=\"%s\">" % self.url

    async def get_session(self):
        if self._session is None:
            self._session = aiohttp.ClientSession(
                connector=aiohttp.TCPConnector(force_close=True),
                timeout=aiohttp.ClientTimeout(total=None),
                headers={"X-API-KEY": self.key})
        return self._session

    async def _request(self, method, path, data=None, json=None, params=None, chunked=None):
        session = await self.get_session()
        try:
            # logging.warning("requesting: "+self.url+path)
            i = 0
            MAX_RETRY_COUNT = 8
            RETRY_DELAY = 2
            while i < MAX_RETRY_COUNT:
                async with session.request(method,
                                           self.url + path,
                                           data=data,
                                           params=params,
                                           json=json,
                                           chunked=chunked,
                                           ssl=self.ssl_context) as resp:
                    if resp.status != 200:
                        msg = await resp.text()
                        if resp.status > 500:
                            print("API Error: [%d], retrying %d/%d" % (resp.status, i, MAX_RETRY_COUNT))
                            i += 1
                            await asyncio.sleep(RETRY_DELAY)
                            continue  # retry
                        raise errors.ApiError("%s [%d]" % (msg, resp.status))
                    if resp.content_type != 'application/json':
                        body = await resp.text()
                        if body.lower() != "ok":
                            raise errors.ApiError("Invalid node response: %s" % body)
                        else:
                            return None
                    try:
                        # logging.warning("\trequest is done")
                        return await resp.json()
                    except ValueError:
                        raise errors.ApiError("Invalid node response (not json)")

        except ssl.CertificateError as e:
            raise errors.ApiError("the specified certificate authority did not validate this server")

        except aiohttp.ClientError as e:
            raise errors.ApiError("Cannot contact node at [%s]" % self.url) from e

        raise errors.ApiError("API Error [%d]: max retry count exceeded" % resp.status)



from typing import Optional, Dict, Union, List
import json

from .session import BaseSession
from .folder_type import Folder
from .event import Event
from .event import from_json as event_from_json
from joule import errors


class EventStream:
    """
        API EventStream model. See :ref:`sec-node-event-stream-actions` for details on using the API to
        manipulate event streams.

        Parameters:
            name (str): stream name, must be unique in the parent
            description (str): optional field
    """

    def __init__(self, name: str = "", description: str = "", event_fields: Optional[Dict[str, str]] = None):
        self._id = None
        self.name = name
        self.description = description
        self.event_fields = self._validate_event_fields(event_fields)

    @property
    def id(self) -> int:
        if self._id is None:
            raise errors.ApiError("this is a local model with no ID. See API docs")
        return self._id

    @id.setter
    def id(self, value: int):
        self._id = value

    def to_json(self) -> Dict:
        return {
            "id": self._id,
            "name": self.name,
            "description": self.description,
            "event_fields": self.event_fields
        }

    def __repr__(self):
        return "<joule.api.EventStream id=%r name=%r description=%r>" % (
            self._id, self.name, self.description
        )

    def _validate_event_fields(self, fields: Optional[Dict[str, str]]) -> Dict[str, str]:
        # make sure values are either string or numeric
        if fields is None:
            return {}
        for field in fields:
            if fields[field] not in ['string', 'numeric']:
                raise errors.ApiError("invalid event field type, must be numeric or string")
        return fields


def from_json(json) -> EventStream:
    my_stream = EventStream()
    my_stream.id = json['id']
    my_stream.name = json['name']
    my_stream.description = json['description']
    my_stream.event_fields = json['event_fields']
    return my_stream


class EventStreamInfo:
    """
        API EventStreamInfo model. Received from :meth:`Node.event_stream_info` and should not be created directly.


        Parameters:
            start (int): timestamp in UNIX microseconds of the beginning of the first event
            end (int): timestamp in UNIX microsseconds of the end of the last data event
            event_count (int): number of events in the stream
            bytes (int): approximate size of the data on disk
            total_time (int): event stream duration in microseconds (start-end)

        """

    def __init__(self, start: Optional[int], end: Optional[int], event_count: int,
                 total_time: int = 0, bytes: int = 0):
        self.start = start
        self.end = end
        self.event_count = event_count
        self.bytes = bytes
        self.total_time = total_time

    def __repr__(self):
        return "<joule.api.EventStreamInfo start=%r end=%r events=%r, total_time=%r>" % (
            self.start, self.end, self.event_count, self.total_time)


def info_from_json(json) -> EventStreamInfo:
    if json is not None:

        return EventStreamInfo(json['start'],
                               json['end'],
                               json['event_count'],
                               json['total_time'],
                               json['bytes'])
    else:
        return EventStreamInfo(None,
                               None,
                               0,
                               0,
                               0)


async def event_stream_delete(session: BaseSession,
                              stream: Union[EventStream, str, int]) -> None:
    data = {}
    if type(stream) is EventStream:
        data["id"] = stream.id
    elif type(stream) is int:
        data["id"] = stream
    elif type(stream) is str:
        data["path"] = stream
    else:
        raise errors.ApiError("Invalid stream datatype. Must be EventStream, Path, or ID")

    await session.delete("/event.json", data)


async def event_stream_create(session: BaseSession,
                              stream: EventStream, folder: Union[Folder, str, int]) -> EventStream:
    data = {"stream": stream.to_json()}

    if type(folder) is Folder:
        data["dest_id"] = folder.id
    elif type(folder) is int:
        data["dest_id"] = folder
    elif type(folder) is str:
        data["dest_path"] = folder
    else:
        raise errors.ApiError("Invalid folder datatype. Must be Folder, Path, or ID")

    resp = await session.post("/event.json", json=data)
    return from_json(resp)


async def event_stream_info(session: BaseSession,
                            stream: Union[EventStream, str, int]) -> EventStreamInfo:
    data = {}

    if type(stream) is EventStream:
        data["id"] = stream.id
    elif type(stream) is int:
        data["id"] = stream
    elif type(stream) is str:
        data["path"] = stream
    else:
        raise errors.ApiError("Invalid stream datatype. Must be EventStream, Path, or ID")

    resp = await session.get("/event.json", data)
    return info_from_json(resp['data_info'])


async def event_stream_get(session: BaseSession,
                           stream: Union[EventStream, str, int],
                           create: bool = False,
                           description: str = "",
                           event_fields=None) -> EventStream:
    data = {}

    if type(stream) is EventStream:
        data["id"] = stream.id
    elif type(stream) is int:
        data["id"] = stream
    elif type(stream) is str:
        data["path"] = stream
    else:
        raise errors.ApiError("Invalid stream datatype. Must be EventStream, Path, or ID")

    try:
        resp = await session.get("/event.json", data)
    except errors.ApiError as e:
        # pass the error if the stream should not or cannot be created
        if not create or type(stream) is not str:
            raise e
        name = stream.split('/')[-1]
        path = '/'.join(stream.split('/')[:-1])
        event_stream = EventStream(name, description, event_fields)
        return await event_stream_create(session, event_stream, path)
    return from_json(resp)


async def event_stream_update(session: BaseSession,
                              stream: EventStream) -> None:
    await session.put("/event.json", {"id": stream.id,
                                      "stream": stream.to_json()})


async def event_stream_move(session: BaseSession,
                            source: Union[EventStream, str, int],
                            destination: Union[Folder, str, int]) -> None:
    data = {}

    if type(source) is EventStream:
        data["src_id"] = source.id
    elif type(source) is int:
        data["src_id"] = source
    elif type(source) is str:
        data["src_path"] = source
    else:
        raise errors.ApiError("Invalid source datatype. Must be EventStream, Path, or ID")

    if type(destination) is Folder:
        data["dest_id"] = destination.id
    elif type(destination) is int:
        data["dest_id"] = destination
    elif type(destination) is str:
        data["dest_path"] = destination
    else:
        raise errors.ApiError("Invalid destination datatype. Must be Folder, Path, or ID")
    await session.put("/event/move.json", data)


async def event_stream_write(session: BaseSession,
                             stream: Union[EventStream, str, int],
                             events: List[Event]) -> List[Event]:
    data = {}
    if type(stream) is EventStream:
        data["id"] = stream.id
    elif type(stream) is int:
        data["id"] = stream
    elif type(stream) is str:
        data["path"] = stream
    else:
        raise errors.ApiError("Invalid stream datatype. Must be EventStream, Path, or ID")
    # post events in blocks
    rx_events = []
    for idx in range(0, len(events), 500):
        chunk = events[idx:idx + 500]
        data['events'] = [e.to_json() for e in events[idx:idx + 500]]
        resp = await session.post("/event/data.json", data)
        rx_events += [event_from_json(e) for e in resp["events"]]
        # copy the ids over
        for i in range(len(chunk)):
            chunk[i].id = rx_events[i].id
    return events


async def event_stream_read(session: BaseSession,
                            stream: Union[EventStream, str, int],
                            start_time: Optional[int] = None,
                            end_time: Optional[int] = None,
                            limit=None,
                            json_filter=None) -> List[Event]:
    params = {}
    if type(stream) is EventStream:
        params["id"] = stream.id
    elif type(stream) is int:
        params["id"] = stream
    elif type(stream) is str:
        params["path"] = stream
    else:
        raise errors.ApiError("Invalid stream datatype. Must be EventStream, Path, or ID")
    if start_time is not None:
        params['start'] = int(start_time)
    if end_time is not None:
        params['end'] = int(end_time)
    if limit is not None:
        if limit <= 0:
            raise errors.ApiError("Limit must be > 0")
        params['limit'] = limit
        params['return-subset'] = 1
    if json_filter is not None:
        params['filter'] = json_filter
    resp = await session.get("/event/data.json", params)
    return [event_from_json(e) for e in resp["events"]]


async def event_stream_remove(session: BaseSession,
                              stream: Union[EventStream, str, int],
                              start_time: Optional[int] = None,
                              end_time: Optional[int] = None,
                              json_filter=None) -> None:
    params = {}
    if type(stream) is EventStream:
        params["id"] = stream.id
    elif type(stream) is int:
        params["id"] = stream
    elif type(stream) is str:
        params["path"] = stream
    else:
        raise errors.ApiError("Invalid stream datatype. Must be EventStream, Path, or ID")
    if start_time is not None:
        params['start'] = int(start_time)
    if end_time is not None:
        params['end'] = int(end_time)
    if json_filter is not None:
        params['filter'] = json_filter
    await session.delete("/event/data.json", params)



import click
import asyncio
import signal

from joule.cli.config import pass_config
from joule.models.pipes import EmptyPipe

import h5py
import json
from joule import errors
from joule.utilities import human_to_timestamp
from icecream import ic
stop_requested = False


@click.command(name="read")
@click.option('-s', "--start", help="timestamp or descriptive string")
@click.option('-e', "--end", help="timestamp or descriptive string")
@click.option('-l', "--live", is_flag=True, help="subscribe to an active stream")
@click.option('-r', "--max-rows", help="limit response data", type=int)
@click.option('-b', "--show-bounds", is_flag=True, help="include min/max for decimated data")
@click.option('-m', "--mark-intervals", help="include [# interval break] tags", is_flag=True)
@click.option('-i', "--elements", help="only include specified elements (first element is 0)")
@click.option('-f', "--file", help="write output to file in hdf5 format")
@click.argument("stream")
@pass_config
def cmd(config, start, end, live, max_rows, show_bounds, mark_intervals, elements, file, stream):
    """Read data from a stream."""
    signal.signal(signal.SIGTERM, handler)
    signal.signal(signal.SIGINT, handler)

    if live and (start is not None or end is not None):
        raise click.ClickException("specify either --live or --start/--end, not both")
    if live and max_rows is not None:
        raise click.ClickException("cannot specify --live and --max-rows")
    if start is not None:
        try:
            start = human_to_timestamp(start)
        except ValueError:
            raise click.ClickException("invalid start time: [%s]" % start)
    if end is not None:
        try:
            end = human_to_timestamp(end)
        except ValueError:
            raise click.ClickException("invalid end time: [%s]" % end)

    # ---- Element selection ----
    if elements is not None:
        try:
            element_indices = [int(e) for e in elements.split(',')]
        except ValueError:
            raise click.ClickException("specify elements as a comma separated like 1,4,8")
    else:
        element_indices = None

    # ---- HDF5 Output ----
    if file is not None:
        write_to_file = True
    else:
        write_to_file = False
    hdf_file = None

    loop = asyncio.get_event_loop()

    async def _run():
        nonlocal element_indices
        nonlocal hdf_file
        nonlocal start, end
        hdf_data = None
        hdf_timestamps = None
        # progress bar for writing to a file
        bar_ctx = None
        bar = None

        # get the stream object from the API
        stream_obj = await config.node.data_stream_get(stream)
        stream_info = await config.node.data_stream_info(stream)
        if stream_info.rows == 0:
            raise click.ClickException("This stream has no data")
        if live:
            pipe = await config.node.data_subscribe(stream)
        else:
            pipe = await config.node.data_read(stream, start, end, max_rows)
        # find the actual start / end bounds of the data read
        if start is None or start < stream_info.start:
            start = stream_info.start
        if end is None or end > stream_info.end:
            end = stream_info.end
        total_time = end - start
        # make sure the element indices make sense with the actual data type
        if element_indices is not None and (max(element_indices) > pipe.width - 1):
            raise click.ClickException("Maximum element is %d" % (pipe.width - 1))
        if element_indices is None:
            element_indices = list(range(len(stream_obj.elements)))
        while not stop_requested and not pipe.is_empty():
            # get new data from the pipe
            try:
                # read structured data
                sdata = await asyncio.wait_for(pipe.read(), 1)
                pipe.consume(len(sdata))
            except asyncio.TimeoutError:
                # check periodically for Ctrl-C (SIGTERM) even if server is slow
                continue
            # ===== Write to HDF File ======
            if write_to_file:
                data_width = len(element_indices)
                target_indices = element_indices #[0] + [idx + 1 for idx in element_indices]
                if hdf_file is None:
                    # create dataset and populate it with current data
                    hdf_data, hdf_timestamps, hdf_file = _create_hdf_dataset(config, stream_obj, stream,
                                                                             element_indices,
                                                                             file, pipe, data_width,
                                                                             initial_size=len(sdata),
                                                                             )
                    hdf_data[...] = sdata['data'][:, target_indices]
                    hdf_timestamps[...] = sdata['timestamp'][:, None]
                    bar_ctx = click.progressbar(length=total_time, label='reading data')
                    # print("total_time: %d" % total_time)
                    bar = bar_ctx.__enter__()
                    chunk_duration = sdata['timestamp'][-1] - sdata['timestamp'][0]
                    bar.update(chunk_duration)
                    # print(chunk_duration)
                else:
                    # expand dataset, append new data
                    cur_size = len(hdf_data)
                    chunk_duration = sdata['timestamp'][-1] - sdata['timestamp'][0]
                    # print("[%d-%d ==> %d]" % (data[-1, 0], hdf_dataset[-1, 0], cum_time))
                    bar.update(chunk_duration)
                    hdf_data.resize((cur_size + len(sdata), data_width))
                    hdf_data[cur_size:, :] = sdata['data'][:, target_indices]
                    hdf_timestamps.resize((cur_size + len(sdata), 1))
                    hdf_timestamps[cur_size:] = sdata['timestamp'][:,None]
                    # update with the new chunk of time
            # ===== Write to stdout (Terminal) ======
            else:
                ts = sdata['timestamp']
                data = sdata['data']
                if pipe.decimated:
                    if show_bounds:
                        # add the bound info
                        num_elements = len(stream_obj.elements)
                        displayed_cols = element_indices + \
                                         [idx + num_elements for idx in element_indices] + \
                                         [idx + num_elements * 2 for idx in element_indices]
                    else:
                        # suppress the bound info
                        displayed_cols = element_indices
                else:
                    displayed_cols = element_indices
                # print out each line, keeping timestamps as integers
                for i in range(len(data)):
                    row = data[i]
                    selected_data = row[displayed_cols]
                    line = "%d %s" % (ts[i], ' '.join('%f' % x for x in selected_data))
                    click.echo(line)
                if pipe.end_of_interval and mark_intervals:
                    click.echo("# interval break")

        await pipe.close()
        if bar_ctx is not None:
            bar.update(end - hdf_timestamps[-1])
            bar_ctx.__exit__(None, None, None)
        if hdf_data is not None:
            hdf_file.close()
        if hdf_timestamps is not None:
            hdf_file.close()

    try:
        loop.run_until_complete(_run())
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


def handler(signum, frame):
    global stop_requested
    stop_requested = True


def _create_hdf_dataset(config, stream, path, element_indices, file, pipe, data_width, initial_size):
    hdf_root = h5py.File(file, "w")
    # note this could be optimized to store data in the correct datatype
    # right now everything is stored as a double which is probably excessive precision
    # but is needed for the timestamps
    hdf_data = hdf_root.create_dataset('data', (initial_size, data_width),
                                       maxshape=(None, data_width), dtype=pipe.dtype[1].base,
                                       compression='gzip')
    hdf_timestamps = hdf_root.create_dataset('timestamp', (initial_size, 1),
                                             maxshape=(None, 1), dtype='i8',
                                             compression='gzip')
    elements = stream.elements
    if element_indices is not None:
        selected_elements = [elements[idx] for idx in element_indices]
    else:
        selected_elements = elements
    element_json = [e.to_json() for e in selected_elements]
    hdf_root.attrs['node_name'] = config.node.name
    hdf_root.attrs['node_url'] = config.node.url
    hdf_root.attrs['path'] = path
    hdf_root.attrs['elements'] = json.dumps([e.name for e in selected_elements])
    hdf_root.attrs['element_json'] = json.dumps(element_json)
    return hdf_data, hdf_timestamps, hdf_root



"""
Configuration data structure for joule
Use load_configs to retrieve Configs object
"""

import enum
from joule.models import Proxy
from typing import Optional, List
import ssl

DEFAULT_CONFIG = {
    "Main":
        {
            "Name": "joule_node",
            "ModuleDirectory": "/etc/joule/module_configs",
            "StreamDirectory": "/etc/joule/stream_configs",
            "SocketDirectory": "/tmp/joule",
            "Database": "joule@localhost:5432/joule",
            "InsertPeriod": 5,
            "CleanupPeriod": 60,
            "MaxLogLines": 100,
        },
    "Proxies": {}
}


class SecurityConfig:
    def __init__(self,
                 certfile: str,
                 keyfile: str,
                 cafile: str):
        self.certfile = certfile
        self.keyfile = keyfile
        self.cafile = cafile


class JouleConfig:
    def __init__(self,
                 name: str,
                 module_directory: str,
                 stream_directory: str,
                 ip_address: Optional[str],
                 port: Optional[int],
                 socket_directory: str,
                 database: str,
                 insert_period: int,
                 cleanup_period: int,
                 max_log_lines: int,
                 nilmdb_url: Optional[str],
                 users_file: Optional[str],
                 proxies: List[Proxy],
                 security: Optional[SecurityConfig]):
        self.name = name
        self.module_directory = module_directory
        self.stream_directory = stream_directory
        self.ip_address = ip_address
        self.port = port
        self.database = database
        self.insert_period = insert_period
        self.cleanup_period = cleanup_period
        self.max_log_lings = max_log_lines
        self.nilmdb_url = nilmdb_url
        self.proxies = proxies
        self.security = security
        self.socket_directory = socket_directory
        self.users_file = users_file



import numpy as np
import asyncio
from typing import List, Union, Tuple, Optional, Callable, Coroutine, TYPE_CHECKING, Dict
from abc import ABC, abstractmethod

from joule.models import pipes

if TYPE_CHECKING:
    from joule.models import DataStream

Loop = asyncio.AbstractEventLoop
# starting and ending timestamps
Interval = Tuple[int, int]

Data = Union[Interval, np.array]


class StreamInfo:
    def __init__(self, start: Optional[int], end: Optional[int], rows: int,
                 total_time: int = 0, bytes: int = 0):
        self.start = start
        self.end = end
        self.rows = rows
        self.bytes = bytes
        self.total_time = total_time

    def __repr__(self):
        return "<DataStreamInfo start=%r end=%r rows=%r, total_time=%r>" % (
            self.start, self.end, self.rows, self.total_time)

    def to_json(self):
        return {
            "start": self.start,
            "end": self.end,
            "rows": self.rows,
            "bytes": self.bytes,
            "total_time": self.total_time
        }
    

class DbInfo:
    def __init__(self, path: str, other: int, reserved: int, free: int, size: int):
        self.path = path
        self.other = other
        self.reserved = reserved
        self.free = free
        self.size = size

    def to_json(self):
        return {
            "path": self.path,
            "other": self.other,
            "reserved": self.reserved,
            "free": self.free,
            "size": self.size
        }


class DataStore(ABC):  # pragma: no cover

    @abstractmethod
    async def initialize(self, streams: List['DataStream']):
        pass

    @abstractmethod
    async def insert(self, stream: 'DataStream',
                     data: np.ndarray, start: int, end: int):
        pass

    @abstractmethod
    async def consolidate(self, stream: 'DataStream', start: Optional[int], end: Optional[int], max_gap: int) -> int:
        pass

    @abstractmethod
    async def spawn_inserter(self, stream: 'DataStream', pipe: pipes.Pipe, insert_period=None) -> asyncio.Task:
        pass

    @abstractmethod
    async def intervals(self, stream: 'DataStream', start: Optional[int], end: Optional[int]):
        pass

    @abstractmethod
    async def extract(self, stream: 'DataStream', start: Optional[int], end: Optional[int],
                      callback: Callable[[np.ndarray, str, bool], Coroutine],
                      max_rows: int = None, decimation_level=None):
        pass

    @abstractmethod
    async def remove(self, stream: 'DataStream', start: Optional[int], end: Optional[int]):
        pass

    @abstractmethod
    async def destroy(self, stream: 'DataStream'):
        pass

    @abstractmethod
    async def destroy_all(self):
        pass

    @abstractmethod
    async def info(self, streams: List['DataStream']) -> Dict[int, StreamInfo]:
        pass

    @abstractmethod
    async def dbinfo(self) -> DbInfo:
        pass

    @abstractmethod
    async def close(self):
        pass



import asyncio


def reader_factory(fd):

    async def f():
        reader = asyncio.StreamReader()
        reader_protocol = asyncio.StreamReaderProtocol(reader)
        src = open(fd, 'rb', 0)
        loop = asyncio.get_running_loop()
        (transport, _) = await loop.connect_read_pipe(
            lambda
            : reader_protocol, src)

        def close_cb():
            transport.close()

        return reader, close_cb

    return f


def writer_factory(fd):

    async def f():
        write_protocol = asyncio.StreamReaderProtocol(asyncio.StreamReader())
        dest = open(fd, 'wb', 0)
        loop = asyncio.get_running_loop()
        (transport, p) = await loop.connect_write_pipe(
            lambda: write_protocol, dest)
        writer = asyncio.StreamWriter(
            transport, write_protocol, None, loop)
        return writer

    return f



import click
import asyncio

from joule import errors
from joule.cli.config import Config, pass_config
from joule.utilities import timestamp_to_human, human_to_timestamp


@click.command(name="consolidate")
@click.option('-s', "--start", help="timestamp or descriptive string")
@click.option('-e', "--end", help="timestamp or descriptive string")
@click.option('-m', "--max-gap", help="remove intervals shorter than this (in us)", default=2e6)
@click.argument("stream")
@pass_config
def consolidate(config: Config, start, end, max_gap, stream: str):
    """Remove gaps in a data stream."""
    if start is not None:
        try:
            start = human_to_timestamp(start)
        except ValueError:
            raise click.ClickException("invalid start time: [%s]" % start)
    if end is not None:
        try:
            end = human_to_timestamp(end)
        except ValueError:
            raise click.ClickException("invalid end time: [%s]" % end)

    loop = asyncio.get_event_loop()
    try:
        num_removed = loop.run_until_complete(config.node.data_consolidate(stream, max_gap, start, end))
        if num_removed > 0:
            print("Consolidated %d intervals" % num_removed)
        else:
            print("No intervals less than %d us" % max_gap)
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    except Exception as e:
        import traceback
        traceback.print_exc()
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()



from aiohttp import web
from sqlalchemy.orm import Session, exc
import secrets
from typing import Dict

from joule.api.session import TcpSession
from joule.models.master import Master, make_key
from joule.utilities.misc import detect_url
from joule import errors


# NOTE: This controller is best tested with e2e
# ignored for unittests

async def add(request: web.Request):  # pragma: no cover
    """
    Contact specified node and add self as a follower
    """
    db: Session = request.app["db"]
    local_name: str = request.app["name"]
    local_port: int = request.app["port"]
    local_scheme: str = request.app["scheme"]
    local_uri: str = request.app["base_uri"]
    if request.content_type != 'application/json':
        return web.Response(text='content-type must be application/json', status=400)
    body = await request.json()

    try:
        master_type = body['master_type'].lower()
        identifier = body['identifier'].lower()
        if master_type not in ['user', 'joule', 'lumen']:
            return web.Response(text="master_type must be [user|joule|lumen]", status=400)
    except (KeyError, ValueError) as e:
        return web.Response(text="Invalid request %s" % str(e), status=400)
    my_master = None
    try:
        if master_type == 'user':
            # check if this name is associated with a master entry
            my_master = db.query(Master). \
                filter(Master.type == Master.TYPE.USER). \
                filter(Master.name == identifier).first()
            if my_master is not None:
                return web.Response(text="User already exists", status=400)
            # see if API key was included in request
            if 'api_key' in body and body['api_key'] is not None:
                key = body['api_key']
                if len(key) < 32:
                    return web.Response(text="API Key must be at least 32 characters", status=400)
            else:
                key = make_key()  # make a new key
            # create a new user API key
            my_master = Master(name=identifier,
                               type=Master.TYPE.USER,
                               key=key)
            grantor_key = request.headers["X-API-KEY"]
            my_master.grantor_id = db.query(Master.id).filter(
                Master.key == grantor_key).one()[0]
            db.add(my_master)
            db.commit()
            return web.json_response({"key": my_master.key,
                                      "name": local_name})
        elif master_type == 'joule' or master_type == 'lumen':
            # create a new node API key
            my_master = Master()
            my_master.key = make_key()
            if master_type == 'joule':
                my_master.type = Master.TYPE.JOULE_NODE
            else:
                my_master.type = Master.TYPE.LUMEN_NODE
            grantor_key = request.headers["X-API-KEY"]

            my_master.grantor_id = db.query(Master.id).filter(
                Master.key == grantor_key).one()[0]
            # make a temporary name
            my_master.name = "pending_%s" % secrets.token_hex(5)
            db.add(my_master)
            db.commit()
            # post this key to the master
            if master_type == 'joule':
                coro = _send_node_key(my_master.key, identifier, local_port,
                                      local_name, local_scheme, local_uri, request.app["cafile"])
            else:
                coro = _send_lumen_key(my_master.key, identifier, local_port,
                                       local_name, local_scheme, local_uri, request.app["cafile"],
                                       body['lumen_params'])
            try:
                name = await coro
            except errors.ApiError as e:
                # remove the pending key
                db.delete(my_master)
                db.commit()
                return web.Response(text="The master cannot add this node [%s]" % str(e), status=400)

            # remove the key for this node if it already exists
            db.query(Master). \
                filter(Master.type == my_master.type). \
                filter(Master.name == name).delete()
            my_master.name = name
            db.add(my_master)
            db.commit()
            return web.json_response({"name": my_master.name})
    except (ValueError, KeyError) as e:
        if my_master is not None:
            db.delete(my_master)
            db.commit()
        return web.Response(text="Cannot add master: [%s]" % str(e), status=400)


async def _send_node_key(key: str,
                         identifier: str,
                         local_port: int,
                         local_name: str,
                         local_scheme: str,
                         local_uri: str,
                         cafile: str) -> str:  # pragma: no cover
    # if the identifier is an IP address or a domain name, turn it into a URL
    if not identifier.startswith("http"):
        url = await detect_url(identifier, 443)
        if url is None:
            raise errors.ApiError("cannot connect to [%s] on port 443" % identifier)
    else:
        url = identifier

    session = TcpSession(url, key, cafile)
    params = {'key': key,
              'port': local_port,
              'name': local_name,
              'base_uri': local_uri,
              'scheme': local_scheme}
    if cafile != "":
        params["name_is_host"] = 1
    resp = await session.post("/follower.json", json=params)
    await session.close()
    return resp['name']


async def _send_lumen_key(key: str,
                          identifier: str,
                          local_port: int,
                          local_name: str,
                          local_scheme: str,
                          local_uri: str,
                          cafile: str,
                          lumen_params: Dict) -> str:  # pragma: no cover
    # if the identifier is an IP address or a domain name, turn it into a URL
    if not identifier.startswith("http"):
        url = await detect_url(identifier + "/lumen")
        if url is None:
            raise errors.ApiError("cannot connect to [%s] on port 80 or 443" % identifier)
    else:
        url = identifier

    # no key needed to access the lumen /nilms end point
    session = TcpSession(url, "stub_key", cafile)
    params = {'api_key': key,
              'port': local_port,
              'name': local_name,
              'base_uri': local_uri,
              'scheme': local_scheme}
    if lumen_params is not None:
        params = {**params, **lumen_params}
    # TODO enable this for CA checking on Lumen server
    # if cafile != "":
    #    params["name_is_host"] = 1
    try:
        await session.post("/nilms.json", json=params)
    except errors.ApiError as e:
        raise e
    finally:
        await session.close()
    return identifier


async def index(request: web.Request):
    db: Session = request.app["db"]
    masters = db.query(Master).all()
    return web.json_response([m.to_json() for m in masters])


async def delete(request: web.Request):
    """
    Remove the specified node from the list of masters
    """
    db: Session = request.app["db"]
    node_name: str = request.app["name"]
    # to eliminate uninitialized warnings
    str_master_type = ""
    name = ""
    try:
        name = request.query["name"]
        str_master_type = request.query["master_type"]
        if str_master_type != "user":
            str_master_type += "_node"
        master_type = Master.TYPE[str_master_type.upper()]
        master = db.query(Master). \
            filter(Master.name == name). \
            filter(Master.type == master_type). \
            one()
        if master.key == request.headers["X-API-KEY"]:
            return web.Response(text="cannot delete yourself, this would lock you out of the node", status=400)

        db.delete(master)
    except (KeyError, ValueError):
        return web.Response(text="specify name and master_type", status=400)
    except exc.NoResultFound:
        return web.Response(text="%s [%s] is not a master of node [%s]" % (str_master_type, name, node_name),
                            status=404)
    db.commit()
    return web.Response(text="ok")



import click
from .info import cli_info
from .move import cli_move
from .destroy import cli_delete
from .rename import cli_rename
from .copy import cli_copy


@click.group(name="event")
def events():
    """Manage event streams."""
    pass  # pragma: no cover


events.add_command(cli_info)
events.add_command(cli_move)
events.add_command(cli_delete)
events.add_command(cli_rename)
events.add_command(cli_copy)



from typing import List, TYPE_CHECKING, Dict, Optional

from joule import errors
from .session import BaseSession


class Master:
    """
    API Master model. See :ref:`sec-node-master-actions` for details on using the API to
    manage master users and nodes.

    Parameters:
       master_type (str): one of [user|joule_node|lumen_node]
       name: unique identifier for the master (username, node name, or URL)
    """

    def __init__(self, master_type: str, name: str, key: str):
        self.master_type = master_type
        self.name = name
        self.key = key


def from_json(json) -> Master:
    return Master(json['type'], json['name'], "omitted")


async def master_add(session: BaseSession, master_type: str,
                     identifier: str, parameters: Optional[Dict] = None,
                     api_key: Optional[str] = None) -> Master:
    data = {"master_type": master_type,
            "identifier": identifier,
            "lumen_params": parameters,
            "api_key": api_key}
    resp = await session.post("/master.json", json=data)
    if master_type == "user":
        return Master(master_type, resp["name"], resp["key"])
    else:
        return Master(master_type, resp["name"], "omitted")


async def master_delete(session: BaseSession, master_type: str, name: str):
    data = {"master_type": master_type,
            "name": name}
    await session.delete("/master.json", params=data)


async def master_list(session: BaseSession) -> List[Master]:
    resp = await session.get("/masters.json")
    return [from_json(item) for item in resp]
